#include <stdint.h>
#include <math.h>

#include <WiFi.h>
#include <WiFiUDP.h>
#include <SPIFFS.h>

#define LED_RGB 8
#define SCK 4
#define RCK 21
#define SER 22
#define DIGIT1 27
#define DIGIT2 26
#define DIGIT3 25
#define DIGIT4 33
#define TACH0 2
#define TACH1 32
#define BTN1 34
#define BTN2 35
#define BTN3 36
#define BTN4 0

// mapping between bits and segments
// 0 -> G
// 1 -> F
// 2 -> A
// 3 -> B
// 4 -> E
// 5 -> D
// 6 -> C
// 7 -> DP
const uint8_t charmap[] = {
  0b11111100, // 0
  0b01100000, // 1
  0b11011010, // 2
  0b11110010, // 3
  0b01100110, // 4
  0b10110110, // 5
  0b10111110, // 6
  0b11100000, // 7
  0b11111110, // 8
  0b11110110, // 9
  0b11101110, // A
  0b00111110, // b
  0b10011100, // C
  0b01111010, // d
  0b10011110, // E
  0b10001110, // F
  0b00000000, // blank
};

enum FxType {
  FX_NONE,
  FX_WAIT,
  FX_REV,
  FX_FAIL,
};

struct ACHandshake {
  int identifier;
  int version;
  int operationId;
};

enum ACOperation {
  ACOP_HANDSHAKE = 0,
  ACOP_SUBSCRIBE_UPDATE = 1,
  ACOP_SUBSCRIBE_SPOT = 2,
  ACOP_DISMISS = 3,
};

struct ACHandshakeResponse {
  uint16_t carName[50];
  uint16_t driverName[50];
  int identifier;
  int version;
  uint16_t trackName[50];
  uint16_t trackConfig[50];
};

struct RTCarInfo {
  uint16_t identifier;
  int size;
  
  float speed_Kmh;
  float speed_Mph;
  float speed_Ms;
  
  bool isAbsEnabled;
  bool isAbsInAction;
  bool isTcInAction;
  bool isTcEnabled;
  bool isInPit;
  bool isEngineLimiterOn;
  
  float accG_vertical;
  float accG_horizontal;
  float accG_frontal;
  
  int lapTime;
  int lastLap;
  int bestLap;
  int lapCount;
  
  float gas;
  float brake;
  float clutch;
  float engineRPM;
  float steer;
  int gear;
  float cgHeight;
  
  float wheelAngularSpeed[4];
  float slipAngle[4];
  float slipAngle_ContactPatch[4];
  float slipRatio[4];
  float tyreSlip[4];
  float ndSlip[4];
  float load[4];
  float Dy[4];
  float Mz[4];
  float tyreDirtyLevel[4];
  
  float camberRAD[4];
  float tyreRadius[4];
  float tyreLoadedRadius[4];
  
  float suspensionHeight[4];
  
  float carPositionNormalized;
  
  float carSlope;
  
  float carCoordinates[3];
};

struct RTLap {
  int carIdentifierNumber;
  int lap;
  uint16_t driverName[50];
  uint16_t carName[50];
  int time;
};

struct CarSettings {
  int shiftPoint;
};

WiFiUDP udp;

uint8_t ledbits[6];
FxType fx = FX_NONE;
bool fakeit = false;
volatile uint8_t buttonState;
volatile uint8_t buttonEvent;
unsigned long lastChange = 0;
String carName;
CarSettings carSettings;

String lazyWideStringParse(const uint16_t *s, size_t max_len) {
  uint8_t *bytes = (uint8_t *)s;
  String out;
  int i = 0;
  while (i < max_len && bytes[i] != 0 && bytes[i] != '%') {
    out += (char)bytes[i];
    i += 2;
  }
  return out;
}

void toggle(int pin) {
  digitalWrite(pin, HIGH);
  //delayMicroseconds(1);
  digitalWrite(pin, LOW);
  //delayMicroseconds(1);
}

void shift_byte(uint8_t b) {
  for (char i = 0; i < 8; i++) {
    digitalWrite(SER, (b >> i) & 1);
    toggle(SCK);
  }
}

void dispNumber(int n) {
  char buf[5];
  snprintf(buf, 5, "%4d", n);
  for (int j = 0; j < 4; j++) {
    if (buf[j] == 0x20) {
      ledbits[3 - j] = 16;
    } else {
      ledbits[3 - j] = buf[j] - 0x30;
    }
  }
}

void hcf() {
  fx = FX_FAIL;
  while (1) {
    delay(1000);
  }
}

void setup() {
  pinMode(SCK, OUTPUT);
  pinMode(RCK, OUTPUT);
  pinMode(SER, OUTPUT);
  pinMode(DIGIT1, OUTPUT);
  pinMode(DIGIT2, OUTPUT);
  pinMode(DIGIT3, OUTPUT);
  pinMode(DIGIT4, OUTPUT);
  pinMode(TACH0, OUTPUT);
  pinMode(TACH1, OUTPUT);
  pinMode(BTN1, INPUT);
  pinMode(BTN2, INPUT);
  pinMode(BTN3, INPUT);
  pinMode(BTN4, INPUT_PULLUP);
  
  digitalWrite(SCK, LOW);
  
  digitalWrite(RCK, LOW);
  toggle(RCK);

  digitalWrite(SER, LOW);

  digitalWrite(DIGIT1, LOW);
  digitalWrite(DIGIT2, LOW);
  digitalWrite(DIGIT3, LOW);
  digitalWrite(DIGIT4, LOW);

  pinMode(LED_RGB, OUTPUT);

  // initialize LED display
  xTaskCreatePinnedToCore(ledDriver, "LEDDriver", 1024, NULL, 1, NULL, 1);
  xTaskCreate(fxThread, "FX", 1024, NULL, 0, NULL);

  // initialize buttons
  attachInterrupt(digitalPinToInterrupt(BTN1), buttonHandler, CHANGE);
  attachInterrupt(digitalPinToInterrupt(BTN2), buttonHandler, CHANGE);
  attachInterrupt(digitalPinToInterrupt(BTN3), buttonHandler, CHANGE);
  attachInterrupt(digitalPinToInterrupt(BTN4), buttonHandler, CHANGE);
  
  fx = FX_WAIT;

  // initialize console
  Serial.begin(115200);

  // initialize filesystem
  if (!SPIFFS.begin()) {
    Serial.println("Formatting empty SPIFFS partition");
    if (SPIFFS.format()) {
      Serial.println("Formatted successfully");
    } else {
      Serial.println("Error formatting SPIFFS partition");
      hcf();
    }
  } else {
    Serial.println("SPIFFS partition mounted successfully");
  }

  Serial.print("Connecting...");
  // initialize networking
  WiFi.begin("bx", "ohti1Obai@ngah");

  while (WiFi.status() != WL_CONNECTED) {
    delay(100);
    Serial.write(".");
  }

  IPAddress ipaddr = WiFi.localIP();
  Serial.println(" connected");
  Serial.print("IP addr: ");
  Serial.println(ipaddr);

  fx = FX_NONE;
  memset(ledbits, 0, 6);

  for (int i = 0; i < 4; i++) {
    dispNumber(ipaddr[i]);
    delay(1000);
  }

  // set up UDP networking
  if (!udp.begin(4242)) {
    Serial.println("Could not create UDP server");
  }

  handshake();
}

void handshake() {
  ACHandshake h;
  bool handshakeComplete = false;
  
  fx = FX_WAIT;
  // clear digits
  memset(ledbits, 16, 4);
  
  while (1) {
    // send handshake
    if (!udp.beginPacket("192.168.2.31", 9996)) {
      Serial.println("Error creating packet");
    }
    h = { 1, 1, ACOP_HANDSHAKE };
    udp.write((uint8_t *)&h, sizeof(ACHandshake));
    if (!udp.endPacket()) {
      Serial.println("Error sending packet");
    }
  
    // wait for handshake response
    int timeout = 100;
    while (timeout > 0 && udp.parsePacket() == 0) {
      if (getButtonEvent() == 4) {
        Serial.println("Fakin' it!");
        fakeit = true;
        goto handshake_end;
      }
      delay(100);
      timeout--;
    }
    if (timeout > 0) {
      break;
    }
  }

handshake_end:
  if (fakeit) {
    carName = "fake_car";
  } else {
    // parse and print handshake response
    size_t len = udp.available();
    if (len < sizeof(ACHandshakeResponse)) {
      Serial.println("Short AC Handshake response");
      hcf();
    }
    ACHandshakeResponse r;
    udp.read((uint8_t *)&r, len);
  
    Serial.print("Car: ");
    carName = lazyWideStringParse(r.carName, 100);
    Serial.println(carName);
    Serial.print("Driver: ");
    Serial.println(lazyWideStringParse(r.driverName, 100));
    Serial.print("Track: ");
    Serial.print(lazyWideStringParse(r.trackName, 100));
    Serial.print("(");
    Serial.print(lazyWideStringParse(r.trackConfig, 100));
    Serial.println(")");
  
    // send subscription
    if (!udp.beginPacket("192.168.2.31", 9996)) {
      Serial.println("Error creating packet");
    }
    h = { 1, 1, ACOP_SUBSCRIBE_UPDATE };
    udp.write((uint8_t *)&h, sizeof(ACHandshake));
    if (!udp.endPacket()) {
      Serial.println("Error sending packet");
    }
  }

  // load persistent data for car

  String carPath = "/" + carName;
  File f = SPIFFS.open(carPath.c_str(), "r");
  if (f) {
    size_t n = f.read((uint8_t *)&carSettings, sizeof(struct CarSettings));
    if (n < sizeof(struct CarSettings)) {
      Serial.println("Short read in car settings. Loading defaults.");
      carSettings = { 7000 };
    } else {
      Serial.println("Loaded car settings for " + carName);
    }
  } else {
    Serial.println("Loaded default settings for " + carName);
    carSettings = { 7000 };
  }

  fx = FX_NONE;
  memset(ledbits, 0, 6);
}

void saveCarSettings() {
  String carPath = "/" + carName;
  File f = SPIFFS.open(carPath.c_str(), "w");
  if (!f) {
    Serial.println("Could not save car settings");
    return;
  }

  size_t n = f.write((uint8_t *)&carSettings, sizeof(struct CarSettings));
  if (n < sizeof(struct CarSettings)) {
    Serial.println("Short write in car settings.");
    return;
  }

  Serial.println("Saved settings for " + carName);
}

#define DIGIT_DWELL_US 500

#define DISP_RPM 0
#define DISP_KMH 1
#define DISP_MPH 2
#define DISP_N_MODES 3

int dispMode = DISP_RPM;
int fakeRPM = 4242;
unsigned long lastPacket = 0;

void loop() {
  RTCarInfo r;
  
  if (fakeit) {
    r.engineRPM = fakeRPM;
    r.speed_Kmh = 999.9;
    r.speed_Mph = 555.5;
    delay(10);
  } else {
    if (udp.parsePacket() != 0) {
      size_t len = udp.available();
      if (len < sizeof(RTCarInfo)) {
        Serial.println("Short CarInfo");
        hcf();
      }
      udp.read((uint8_t *)&r, len);
      lastPacket = millis();
    } else {
      delay(10);
    }

    if (millis() - lastPacket > 10000) {
      // Ten seconds since last packet - let's retry the handshake
      handshake();
      return;
    }
  }
  
  int rpm = (int) r.engineRPM;
  int d_data;
  switch (dispMode) {
    case DISP_RPM:
      d_data = rpm;
      break;
    case DISP_KMH:
      d_data = (int) r.speed_Kmh;
      break;
    case DISP_MPH:
      d_data = (int) r.speed_Mph;
      break;
  }
  dispNumber(d_data);

  // calculate lower 7 bits for 0 to (shiftPoint - 2000)
  uint16_t t;
  if (rpm < carSettings.shiftPoint - 2000) {
    t = 0x7F00;
    t <<= 7 - (rpm * 7) / (carSettings.shiftPoint - 2000);
  } else if (rpm < carSettings.shiftPoint) {
    t = 0xFFFF;
    t <<= 9 - ((rpm - (carSettings.shiftPoint - 2000)) * 9) / 2000;
  } else {
    bool b = (millis() / 100) % 2 == 0;
    t = b ? 0xFFF0 : 0xFFFF;    
  }
  ledbits[4] = t >> 8;
  ledbits[5] = t & 0xFF;

  uint8_t b = getButtonEvent();
  if (b == 1) {
    if (buttonState == 0b1001) {
      handshake();
      return;
    }

    dispMode = (dispMode + 1) % DISP_N_MODES;
    dispNumber(dispMode);
    delay(500);
  } else if (b == 2 || b == 3) {
    if (buttonState & 0b1000) {
      if (b == 2) {
        fakeRPM -= 100;
      } else {
        fakeRPM += 100;
      }
      return;
    }
    
    if (b == 2) {
      carSettings.shiftPoint -= 100;
    } else {
      carSettings.shiftPoint += 100;
    }
    dispNumber(carSettings.shiftPoint);
    saveCarSettings();
    delay(250);
  }
}

void ledDriver(void *pvParameters) {
  while (1) {
    shift_byte(ledbits[5]);
    //taskENTER_CRITICAL();
    digitalWrite(DIGIT4, LOW);
    toggle(RCK);
    digitalWrite(TACH1, HIGH);
    //taskEXIT_CRITICAL();
    delayMicroseconds(DIGIT_DWELL_US / 2);
    
    taskYIELD();

    shift_byte(ledbits[4]);
    //taskENTER_CRITICAL();
    digitalWrite(TACH1, LOW);
    toggle(RCK);
    digitalWrite(TACH0, HIGH);
    //taskEXIT_CRITICAL();
    delayMicroseconds(DIGIT_DWELL_US / 2);
    
    taskYIELD();

    shift_byte(charmap[ledbits[3]]);
    //taskENTER_CRITICAL();
    digitalWrite(TACH0, LOW);
    toggle(RCK);
    digitalWrite(DIGIT1, HIGH);
    //taskEXIT_CRITICAL();
    delayMicroseconds(DIGIT_DWELL_US);
    
    taskYIELD();

    shift_byte(charmap[ledbits[2]]);
    //taskENTER_CRITICAL();
    digitalWrite(DIGIT1, LOW);
    toggle(RCK);
    digitalWrite(DIGIT2, HIGH);
    //taskEXIT_CRITICAL();
    delayMicroseconds(DIGIT_DWELL_US);
    
    taskYIELD();

    shift_byte(charmap[ledbits[1]]);
    //taskENTER_CRITICAL();
    digitalWrite(DIGIT2, LOW);
    toggle(RCK);
    digitalWrite(DIGIT3, HIGH);
    //taskEXIT_CRITICAL();
    delayMicroseconds(DIGIT_DWELL_US);

    taskYIELD();
    
    shift_byte(charmap[ledbits[0]]);
    //taskENTER_CRITICAL();
    digitalWrite(DIGIT3, LOW);
    toggle(RCK);
    digitalWrite(DIGIT4, HIGH);
    //taskEXIT_CRITICAL();
    delayMicroseconds(DIGIT_DWELL_US);

    taskYIELD();
  }
}

void fxThread(void *pvParameters) {
  int state = 0;
  bool up = true;
  uint16_t t;
  
  while (1) {
    switch (fx) {
    case FX_WAIT:
      if (up) {
        state++;
      } else {
        state--;
      }
      if (state == 15) {
        up = false;
      }
      if (state == 0) {
        up = true;
      }
      t = 1 << state;
      ledbits[4] = t >> 8;
      ledbits[5] = t & 0xFF;
      break;
    case FX_FAIL:
      ledbits[5] = 0;
      if (state & 0x04) {
        ledbits[4] = 0b11111100;
      } else {
        ledbits[4] = 0;
      }
      state++;
      break;
    case FX_REV:
      if (state == 16) {
        ledbits[4] = ledbits[5] = 0;
        state = 0;
      } else {
        ledbits[5] = (ledbits[5] >> 1) | (ledbits[4] << 7);
        ledbits[4] = (ledbits[4] >> 1) | 0x80;
        state++;
      }
      break;
    }
    delay(100);
  }
}

void buttonHandler() {
  uint8_t newButtonState =
    digitalRead(BTN4) << 3 |
    digitalRead(BTN3) << 2 |
    digitalRead(BTN2) << 1 |
    digitalRead(BTN1);
  newButtonState = ~newButtonState & 0xF;

  if (newButtonState != buttonState) {
    unsigned long now = millis();
    long dt = now - lastChange;
    if (dt < 100) return;
    lastChange = now;
    uint8_t pressed = (buttonState ^ newButtonState) & newButtonState;
    switch(pressed) {
      case 0b0001:
        buttonEvent = 1;
        break;
      case 0b0010:
        buttonEvent = 2;
        break;
      case 0b0100:
        buttonEvent = 3;
        break;
      case 0b1000:
        buttonEvent = 4;
        break;
      default:
        buttonEvent = 0;
    }
    buttonState = newButtonState;
  }
}

uint8_t getButtonEvent() {
  if (buttonEvent) {
    uint8_t t = buttonEvent;
    buttonEvent = 0;
    return t;
  }
  return 0;
}
